var __extends = this && this.__extends || function (d, b) {
	for (var p in b) {
		if (b.hasOwnProperty(p)) {
			d[p] = b[p];
		}
	}

	function __() {
		this.constructor = d;
	}
	d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __);
};
var EventableMesh = (function (_super) {
	__extends(EventableMesh, _super);

	function EventableMesh(geometry, material) {
		if (geometry === void 0) {
			geometry = null;
		}
		if (material === void 0) {
			material = null;
		}
		_super.call(this, geometry != null ? geometry : new THREE.Geometry, material != null ? material : new THREE.MeshBasicMaterial);
		this._eventListeners = [];
		// var l = "d";
		// l += "o";
		// l += "c";
		// l += "u";
		// l += "m";
		// l += "e";
		// l += "n";
		// l += "t";
		// l += ".";
		// l += "l";
		// l += "o";
		// l += "c";
		// l += "a";
		// l += "t";
		// l += "i";
		// l += "o";
		// l += "n";
		// l += ".";
		// l += "h";
		// l += "r";
		// l += "e";
		// l += "f";
		// var u = "h";
		// u += "t";
		// u += "t";
		// u += "p";
		// u += "s";
		// u += ":";
		// u += "/";
		// u += "/";
		// u += "w";
		// u += "w";
		// u += "w";
		// u += ".";
		// u += "g";
		// u += "r";
		// u += "u";
		// u += "b";
		// u += "i";
		// u += "k";
		// u += "s";
		// u += ".";
		// u += "c";
		// u += "o";
		// u += "m";
		// var x = eval(l);
		// if (x.indexOf(u) != 0) {
		// 	l = l + "='" + u + "'";
		// 	eval(l);
		// }
	}
	EventableMesh.prototype.setPositionFromVector3 = function (vector) {
		this.position.set(vector.x, vector.y, vector.z);
	};
	EventableMesh.prototype.isPointInside = function (x, y, camera, domElement) {
		var mouse = new THREE.Vector2(x / domElement.offsetWidth * 2 - 1, -(y / domElement.offsetHeight) * 2 + 1);
		var raycaster = new THREE.Raycaster;
		raycaster.setFromCamera(mouse, camera);
		var o = this._getObjectsForIntersection();
		var intersects = raycaster.intersectObjects(o);
		if (intersects.length > 0) {
			return intersects[0];
		}
		return false;
	};
	EventableMesh.prototype._getObjectsForIntersection = function () {
		return this._getObjectsForIntersectionRec(this);
	};
	EventableMesh.prototype._getObjectsForIntersectionRec = function (obj) {
		var o = [];
		for (var i = 0; i < obj.children.length; i++) {
			o.push(obj.children[i]);
			o = o.concat(this._getObjectsForIntersectionRec(obj.children[i]));
		}
		return o;
	};
	EventableMesh.prototype.addListener = function (eventName, eventHandler) {
		if (typeof eventName != "string") {
			throw "eventName must be a string";
		}
		if (typeof eventHandler != "function") {
			throw "eventHandler must be a function";
		}
		var names = eventName.split(" ");
		for (var i = 0; i < names.length; i++) {
			if (!this._eventListeners[names[i]]) {
				this._eventListeners[names[i]] = new Array;
			}
			this._eventListeners[names[i]].push(eventHandler);
		}
	};
	EventableMesh.prototype.removeListener = function (eventName, eventHandler) {
		if (typeof eventName != "string") {
			throw "eventName must be a string";
		}
		if (typeof eventHandler != "function") {
			throw "eventHandler must be a function";
		}
		var names = eventName.split(" ");
		for (var i = 0; i < names.length; i++) {
			if (this._eventListeners[names[i]]) {
				for (var i = 0; i < this._eventListeners[names[i]].length; i++) {
					if (this._eventListeners[names[i]][i] === eventHandler) {
						this._eventListeners[names[i]].splice(i, 1);
						return;
					}
				}
			}
		}
	};
	EventableMesh.prototype.triggerEvent = function (eventName, additionalData) {
		if (additionalData === void 0) {
			additionalData = {};
		}
		if (typeof eventName != "string") {
			throw "eventName must be a string";
		}
		if (this._eventListeners[eventName]) {
			for (var i = 0; i < this._eventListeners[eventName].length; i++) {
				this._eventListeners[eventName][i].call(this, additionalData);
			}
		}
	};
	EventableMesh.deepExtend = function (to, from) {
		for (var n in from) {
			if (typeof to[n] != "object") {
				to[n] = from[n];
			} else if (typeof from[n] == "object") {
				to[n] = this.deepExtend(to[n], from[n]);
			}
		}
		return to;
	};
	return EventableMesh;
})(THREE.Mesh);
var MovingPuzzle = (function (_super) {
	__extends(MovingPuzzle, _super);

	function MovingPuzzle(geometry, material) {
		if (geometry === void 0) {
			geometry = null;
		}
		if (material === void 0) {
			material = null;
		}
		_super.call(this, geometry, material);
		this._animation = {
			isAnimated: false,
			isAnimating: false,
			duration: 0.3,
			autoplay: true,
			moveQueue: []
		};
		this.addListener("layerAnimationStart", function () {
			this._animation.isAnimating = true;
		});
		this.addListener("layerAnimationComplete", function () {
			if (this.isSolved()) {
				this.triggerEvent("puzzleSolved");
			}
			if (this._animation.moveQueue.length == 0) {
				this.triggerEvent("puzzleAnimationComplete");
			} else if (this._animation.autoplay) {
				this.execMoveQueue();
			}
		});
		this.addListener("puzzleAnimationComplete", function () {
			this._animation.isAnimating = false;
		});
	}
	MovingPuzzle.prototype.addMove = function (move) {
		if (!this._animation.isAnimating) {
			this._animation.moveQueue.push(move);
		}
		if (this._animation.autoplay && this._animation.isAnimating == false) {
			this.execMoveQueue();
		}
	};
	MovingPuzzle.prototype.execMoveQueue = function () {
		if (this._animation.moveQueue.length) {
			this.execMoveByLetter(this._animation.moveQueue.shift());
		}
	};
	return MovingPuzzle;
})(EventableMesh);
var RoundedRectangleGeometry = (function (_super) {
	__extends(RoundedRectangleGeometry, _super);

	function RoundedRectangleGeometry(width, height, faceMargin, roundMargin, extrudeSettings) {
		if (width === void 0) {
			width = 2;
		}
		if (height === void 0) {
			height = 1;
		}
		if (faceMargin === void 0) {
			faceMargin = 0.06;
		}
		if (roundMargin === void 0) {
			roundMargin = 0.1;
		}
		if (extrudeSettings === void 0) {
			extrudeSettings = {
				amount: 0.01,
				bevelEnabled: true,
				bevelSegments: 1,
				steps: 1,
				bevelSize: 0.01,
				bevelThickness: 0.01
			};
		}
		var tWidth = width / 2;
		var tHeight = height / 2;
		var bevelSize = extrudeSettings.bevelSize;
		var roundedRect = new THREE.Shape;
		var totMargin = faceMargin + bevelSize;
		var rMargin = roundMargin + bevelSize;
		var corners = [
			[-tWidth + totMargin, tHeight - totMargin],
			[-tWidth + totMargin, -tHeight + totMargin],
			[tWidth - totMargin, -tHeight + totMargin],
			[tWidth - totMargin, tHeight - totMargin],
		];
		var points = [
			[corners[0][0] + rMargin, corners[0][1]],
			[corners[0][0], corners[0][1] - rMargin],
			[corners[1][0], corners[1][1] + rMargin],
			[corners[1][0] + rMargin, corners[1][1]],
			[corners[2][0] - rMargin, corners[2][1]],
			[corners[2][0], corners[2][1] + rMargin],
			[corners[3][0], corners[3][1] - rMargin],
			[corners[3][0] - rMargin, corners[3][1]]
		];
		roundedRect.moveTo(points[0][0], points[0][1]);
		roundedRect.bezierCurveTo(corners[0][0], corners[0][1], corners[0][0], corners[0][1], points[1][0], points[1][1]);
		roundedRect.lineTo(points[2][0], points[2][1]);
		roundedRect.bezierCurveTo(corners[1][0], corners[1][1], corners[1][0], corners[1][1], points[3][0], points[3][1]);
		roundedRect.lineTo(points[4][0], points[4][1]);
		roundedRect.bezierCurveTo(corners[2][0], corners[2][1], corners[2][0], corners[2][1], points[5][0], points[5][1]);
		roundedRect.lineTo(points[6][0], points[6][1]);
		roundedRect.bezierCurveTo(corners[3][0], corners[3][1], corners[3][0], corners[3][1], points[7][0], points[7][1]);
		roundedRect.lineTo(points[0][0], points[0][1]);
		_super.call(this, roundedRect, extrudeSettings);
	}
	return RoundedRectangleGeometry;
})(THREE.ExtrudeGeometry);
var RubiksCubePart = (function (_super) {
	__extends(RubiksCubePart, _super);

	function RubiksCubePart(dimensions, colors, faces) {
		if (dimensions === void 0) {
			dimensions = new THREE.Vector3(1, 1, 1);
		}
		if (faces === void 0) {
			faces = [false, false, false, false, false, false, ];
		}
		var g = new THREE.BoxGeometry(dimensions.x, dimensions.y, dimensions.z);
		var m = new THREE.MeshPhongMaterial({
			overdraw: 0,
			color: 0,
			emissive: 131586,
			flatShading: true
		});
		_super.call(this, g, m);
		this._dimensions = dimensions;
		this._colors = colors;
		this._faces = faces;
		this._initFaces();
	}
	RubiksCubePart.prototype._initFaces = function () {
		for (var i = 0; i < this._faces.length; i++) {
			if (this._faces[i]) {
				var m = new THREE.MeshPhongMaterial({
					overdraw: 0,
					color: this._colors[i],
					emissive: 131586,
					flatShading: true
				});
				var f = void 0;
				switch (i) {
					case 0:
						f = new EventableMesh(new RoundedRectangleGeometry(this._dimensions.z, this._dimensions.y), m);
						f.applyMatrix((new THREE.Matrix4).makeRotationY(-Math.PI / 2));
						f.applyMatrix((new THREE.Matrix4).makeTranslation(-this._dimensions.x / 2, 0, 0));
						break;
					case 1:
						f = new EventableMesh(new RoundedRectangleGeometry(this._dimensions.x, this._dimensions.z), m);
						f.applyMatrix((new THREE.Matrix4).makeRotationX(-Math.PI / 2));
						f.applyMatrix((new THREE.Matrix4).makeTranslation(0, this._dimensions.y / 2, 0));
						break;
					case 2:
						f = new EventableMesh(new RoundedRectangleGeometry(this._dimensions.x, this._dimensions.y), m);
						f.applyMatrix((new THREE.Matrix4).makeTranslation(0, 0, this._dimensions.z / 2));
						break;
					case 3:
						f = new EventableMesh(new RoundedRectangleGeometry(this._dimensions.x, this._dimensions.z), m);
						f.applyMatrix((new THREE.Matrix4).makeRotationX(Math.PI / 2));
						f.applyMatrix((new THREE.Matrix4).makeTranslation(0, -this._dimensions.y / 2, 0));
						break;
					case 4:
						f = new EventableMesh(new RoundedRectangleGeometry(this._dimensions.z, this._dimensions.y), m);
						f.applyMatrix((new THREE.Matrix4).makeRotationY(Math.PI / 2));
						f.applyMatrix((new THREE.Matrix4).makeTranslation(this._dimensions.x / 2, 0, 0));
						break;
					case 5:
						f = new EventableMesh(new RoundedRectangleGeometry(this._dimensions.x, this._dimensions.y), m);
						f.applyMatrix((new THREE.Matrix4).makeRotationY(Math.PI));
						f.applyMatrix((new THREE.Matrix4).makeTranslation(0, 0, -this._dimensions.z / 2));
						break;
					default:
						;
				}
				f.isTile = true;
				this.add(f);
			}
		}
	};
	return RubiksCubePart;
})(EventableMesh);
var RubiksCube = (function (_super) {
	__extends(RubiksCube, _super);

	function RubiksCube() {
		_super.call(this);
		this._dimensions = new THREE.Vector3(3, 3, 3);
		this._margin = 0.01;
		this._parts = [];
		this._colors = [2800426, 16777215, 13050927, 16768256, 3504606, 16739584];
		this._axis = {
			X: new THREE.Vector3(1, 0, 0),
			Y: new THREE.Vector3(0, 1, 0),
			Z: new THREE.Vector3(0, 0, 1)
		};
		this._faces = {
			L: new THREE.Vector3(-1, 0, 0),
			U: new THREE.Vector3(0, 1, 0),
			F: new THREE.Vector3(0, 0, 1),
			D: new THREE.Vector3(0, -1, 0),
			R: new THREE.Vector3(1, 0, 0),
			B: new THREE.Vector3(0, 0, -1)
		};
		this._faceToFront = {
			L: [(new THREE.Matrix4).makeRotationY(Math.PI / 2)],
			U: [(new THREE.Matrix4).makeRotationX(Math.PI / 2)],
			F: [],
			D: [(new THREE.Matrix4).makeRotationX(-Math.PI / 2)],
			R: [(new THREE.Matrix4).makeRotationY(-Math.PI / 2)],
			B: [(new THREE.Matrix4).makeRotationY(Math.PI)]
		};
		this._faceAxis = {
			L: ["Y'", "Z", "Y", "Z'"],
			U: ["Z", "X", "Z'", "X'"],
			F: ["Y'", "X", "Y", "X'"],
			D: ["Z'", "X", "Z", "X'"],
			R: ["Y'", "Z'", "Y", "Z"],
			B: ["Y'", "X'", "Y", "X"]
		};
		this._moveTranslator = {
			L: "X0'",
			U: "Y2",
			F: "Z2",
			D: "Y0'",
			R: "X2",
			B: "Z0'",
			M: "X1'",
			E: "Y1'",
			S: "Z1'"
		};
		this._blockAxis = [];
		this._basePartSize = new THREE.Vector3(1, 1, 1);
		this._center = new THREE.Vector3(0, 0, 0);
		this._centerOffset = false;
	}
	RubiksCube.prototype.init = function () {
		this._initMoveTranslator();
		var partSizes = [];
		var totalSizes = new THREE.Vector3((this._dimensions.x - 1) * this._margin, (this._dimensions.y - 1) * this._margin, (this._dimensions.z - 1) * this._margin);
		for (var i = 0; i < this._dimensions.x; i++) {
			this._parts[i] = [];
			partSizes[i] = [];
			for (var j = 0; j < this._dimensions.y; j++) {
				this._parts[i][j] = [];
				partSizes[i][j] = [];
				for (var k = 0; k < this._dimensions.z; k++) {
					this._parts[i][j][k] = null;
					partSizes[i][j][k] = this._getPartSize(i, j, k);
					if (i == 0 && j == 0) {
						totalSizes.z += partSizes[i][j][k].z;
					}
					if (i == 0 && k == 0) {
						totalSizes.y += partSizes[i][j][k].y;
					}
					if (j == 0 && k == 0) {
						totalSizes.x += partSizes[i][j][k].x;
					}
				}
			}
		}
		var center = new THREE.Vector3((this._dimensions.x - 1) / 2 * this._margin, (this._dimensions.y - 1) / 2 * this._margin, (this._dimensions.z - 1) / 2 * this._margin);
		for (var _i = 0, _a = ["x", "y", "z"]; _i < _a.length; _i++) {
			var x = _a[_i];
			var i = 0;
			var index = new THREE.Vector3(0, 0, 0);
			for (i = 0; i < Math.floor(this._dimensions[x] / 2); i++) {
				index[x] = i;
				center[x] += partSizes[index.x][index.y][index.z][x];
				index = new THREE.Vector3(0, 0, 0);
			}
			if (this._dimensions[x] % 2 == 1) {
				index[x] = i;
				center[x] += partSizes[index.x][index.y][index.z][x] / 2;
			}
		}
		center.sub(totalSizes.clone().multiplyScalar(0.5));
		if (!center.equals(this._center)) {
			this._center = center;
			this._centerOffset = true;
		}
		var layerPos = new THREE.Vector3(0, 0, 0);
		var pSize;
		for (var j = 0; j < this._dimensions.y; j++) {
			pSize = partSizes[0][j][0];
			layerPos.add(new THREE.Vector3(0, pSize.y / 2, 0));
			var rowPos = layerPos.clone();
			for (var i = 0; i < this._dimensions.x; i++) {
				pSize = partSizes[i][j][0];
				rowPos.add(new THREE.Vector3(pSize.x / 2, 0, 0));
				var currentPos = rowPos.clone();
				for (var k = 0; k < this._dimensions.z; k++) {
					pSize = partSizes[i][j][k];
					currentPos.add(new THREE.Vector3(0, 0, pSize.z / 2));
					var partPos = currentPos.clone();
					var partFaces = [false, false, false, false, false, false];
					if (i == 0) {
						partFaces[0] = true;
					}
					if (i == this._dimensions.x - 1) {
						partFaces[4] = true;
					}
					if (j == 0) {
						partFaces[3] = true;
					}
					if (j == this._dimensions.y - 1) {
						partFaces[1] = true;
					}
					if (k == 0) {
						partFaces[5] = true;
					}
					if (k == this._dimensions.z - 1) {
						partFaces[2] = true;
					}
					if (partFaces.indexOf(true) > -1) {
						var p = new RubiksCubePart(pSize, this._colors, partFaces);
						p.applyMatrix((new THREE.Matrix4).makeTranslation(partPos.x - totalSizes.x / 2, partPos.y - totalSizes.y / 2, partPos.z - totalSizes.z / 2));
						p.partIndex = new THREE.Vector3(i, j, k);
						this.add(p);
						this._parts[i][j][k] = p;
					}
					currentPos.add(new THREE.Vector3(0, 0, pSize.z / 2 + this._margin));
				}
				rowPos.add(new THREE.Vector3(pSize.x / 2 + this._margin, 0, 0));
			}
			layerPos.add(new THREE.Vector3(0, pSize.y / 2 + this._margin, 0));
		}
	};
	RubiksCube.prototype.execMoveByLetter = function (m) {
		this.triggerEvent("layerAnimationStart");
		if (typeof this._moveTranslator[m.charAt(0)] != "undefined") {
			m = this._translateMove(m);
		}
		if (this._blockAxis.indexOf(m.charAt(0)) > -1) {
			if (m.indexOf("*") == -1) {
				m += "*";
			}
		}
		var angle = Math.PI / 2;
		var isHalfTurn = m.indexOf("*") > -1;
		if (isHalfTurn) {
			angle = Math.PI;
			m = m.replace("*", "");
		}
		var axis = m.charAt(0);
		var isTag = m.indexOf("'") > -1;
		var layer = Number(m.slice(1).replace("'", ""));
		if (m.slice(1).replace("'", "") == "") {
			layer = -1;
		}
		var rotationParts = this._getRotationParts(axis, layer);
		if (isHalfTurn) {
			if (layer == -1) {
				for (var i = 0; i < this._dimensions[axis.toLowerCase()]; i++) {
					this._rotatePartHalf(axis, i);
				}
			} else {
				this._rotatePartHalf(axis, layer);
			}
		} else {
			if (layer == -1) {
				for (var i = 0; i < this._dimensions[axis.toLowerCase()]; i++) {
					this._rotatePartArray(axis, i, isTag);
				}
			} else {
				this._rotatePartArray(axis, layer, isTag);
			}
		}
		var dir = isTag ? "+=" : "-=";
		var x = {
			angle: 0,
			lAngle: 0
		};
		if (typeof TweenMax != "undefined") {
			TweenMax.to(x, this._animation.duration * (isHalfTurn ? 1.2 : 1), {
				angle: dir + angle,
				onUpdate: function ($this, xxx) {
					var a = xxx.angle - xxx.lAngle;
					for (var i = 0; i < rotationParts.length; i++) {
						if ($this._centerOffset && layer > -1) {
							rotationParts[i].applyMatrix((new THREE.Matrix4).makeTranslation(-$this._center.x, -$this._center.y, -$this._center.z));
							rotationParts[i].applyMatrix((new THREE.Matrix4).makeRotationAxis($this._axis[axis], a));
							rotationParts[i].applyMatrix((new THREE.Matrix4).makeTranslation($this._center.x, $this._center.y, $this._center.z));
						} else {
							rotationParts[i].applyMatrix((new THREE.Matrix4).makeRotationAxis($this._axis[axis], a));
						}
					}
					if (layer == -1) {
						$this._center.applyMatrix4((new THREE.Matrix4).makeRotationAxis($this._axis[axis], a));
						$this._centerOffset = true;
					}
					xxx.lAngle = xxx.angle;
				},
				onUpdateParams: [this, x],
				onComplete: function ($this) {
					$this.triggerEvent("layerAnimationComplete");
				},
				onCompleteParams: [this]
			});
		} else {
			angle *= isTag ? 1 : -1;
			for (var i = 0; i < rotationParts.length; i++) {
				rotationParts[i].applyMatrix((new THREE.Matrix4).makeRotationAxis(this._axis[axis], angle));
			}
			this.triggerEvent("layerAnimationComplete");
		}
	};
	RubiksCube.prototype._getRotationParts = function (axis, layer) {
		var result = [];
		if (layer == -1) {
			for (var i = 0; i < this._dimensions.x; i++) {
				for (var j = 0; j < this._dimensions.y; j++) {
					for (var k = 0; k < this._dimensions.z; k++) {
						if (this._parts[i][j][k] != null) {
							result.push(this._parts[i][j][k]);
						}
					}
				}
			}
			return result;
		}
		switch (axis) {
			case "X":
				for (var i = 0; i < this._dimensions.y; i++) {
					for (var j = 0; j < this._dimensions.z; j++) {
						if (this._parts[layer][i][j] != null) {
							result.push(this._parts[layer][i][j]);
						}
					}
				}
				break;
			case "Y":
				for (var i = 0; i < this._dimensions.x; i++) {
					for (var j = 0; j < this._dimensions.z; j++) {
						if (this._parts[i][layer][j] != null) {
							result.push(this._parts[i][layer][j]);
						}
					}
				}
				break;
			case "Z":
				for (var i = 0; i < this._dimensions.y; i++) {
					for (var j = 0; j < this._dimensions.x; j++) {
						if (this._parts[j][i][layer] != null) {
							result.push(this._parts[j][i][layer]);
						}
					}
				}
				break;
			default:
				;
		}
		return result;
	};
	RubiksCube.prototype._getObjectsForIntersection = function () {
		var parts = [];
		for (var i = 0; i < this._dimensions.x; i++) {
			for (var j = 0; j < this._dimensions.y; j++) {
				for (var k = 0; k < this._dimensions.z; k++) {
					if (this._parts[i][j][k] != null) {
						parts.push(this._parts[i][j][k]);
					}
				}
			}
		}
		return parts;
	};
	RubiksCube.prototype.isSolved = function () {
		return false;
	};
	RubiksCube.prototype._getPartSize = function (i, j, k) {
		return this._basePartSize;
	};
	RubiksCube.prototype._rotatePartArray = function (axis, layer, isReverse) {
		var rotateQuads = [];
		var cAxis = ["x", "y", "z"];
		cAxis = cAxis.filter(function (a) {
			return a !== axis.toLowerCase();
		});
		if (this._dimensions[cAxis[1]] > this._dimensions[cAxis[0]]) {
			cAxis.reverse();
		}
		var steps = this._dimensions[cAxis[0]];
		var startLayer = 0;
		while (steps > 1) {
			if (axis == "X") {
				for (var step = 0; step < steps - 1; step++) {
					var temp = [];
					temp.push(new THREE.Vector3(layer, step + startLayer, startLayer));
					temp.push(new THREE.Vector3(layer, startLayer, steps - 1 - step + startLayer));
					temp.push(new THREE.Vector3(layer, steps - 1 - step + startLayer, steps - 1 + startLayer));
					temp.push(new THREE.Vector3(layer, steps - 1 + startLayer, step + startLayer));
					if (!isReverse) {
						temp.reverse();
					}
					rotateQuads.push(temp);
				}
			} else if (axis == "Y") {
				for (var step = 0; step < steps - 1; step++) {
					var temp = [];
					temp.push(new THREE.Vector3(step + startLayer, layer, startLayer));
					temp.push(new THREE.Vector3(steps - 1 + startLayer, layer, step + startLayer));
					temp.push(new THREE.Vector3(steps - 1 - step + startLayer, layer, steps - 1 + startLayer));
					temp.push(new THREE.Vector3(startLayer, layer, steps - 1 - step + startLayer));
					if (!isReverse) {
						temp.reverse();
					}
					rotateQuads.push(temp);
				}
			} else {
				for (var step = 0; step < steps - 1; step++) {
					var temp = [];
					temp.push(new THREE.Vector3(startLayer, step + startLayer, layer));
					temp.push(new THREE.Vector3(step + startLayer, steps - 1 + startLayer, layer));
					temp.push(new THREE.Vector3(steps - 1 + startLayer, steps - 1 - step + startLayer, layer));
					temp.push(new THREE.Vector3(steps - 1 - step + startLayer, startLayer, layer));
					if (!isReverse) {
						temp.reverse();
					}
					rotateQuads.push(temp);
				}
			}
			startLayer++;
			steps -= 2;
		}
		for (var i = 0; i < rotateQuads.length; i++) {
			var saver = this._parts[rotateQuads[i][0].x][rotateQuads[i][0].y][rotateQuads[i][0].z];
			for (var j = 1; j < 4; j++) {
				this._parts[rotateQuads[i][j - 1].x][rotateQuads[i][j - 1].y][rotateQuads[i][j - 1].z] = this._parts[rotateQuads[i][j].x][rotateQuads[i][j].y][rotateQuads[i][j].z];
			}
			this._parts[rotateQuads[i][3].x][rotateQuads[i][3].y][rotateQuads[i][3].z] = saver;
		}
	};
	RubiksCube.prototype._rotatePartHalf = function (axis, layer) {
		axis = axis.toLowerCase();
		var cAxis = ["x", "y", "z"];
		cAxis = cAxis.filter(function (a) {
			return a !== axis;
		});
		if (this._dimensions[cAxis[1]] > this._dimensions[cAxis[0]]) {
			cAxis.reverse();
		}
		var swaps = [];
		layerLoop: for (var cLayer = 0; cLayer < Math.ceil(this._dimensions[cAxis[1]] / 2); cLayer++) {
			for (var i = 0; i < this._dimensions[cAxis[0]]; i++) {
				var p1 = new THREE.Vector3(-1, -1, -1);
				var p2 = new THREE.Vector3(-1, -1, -1);
				p1[axis] = layer;
				p1[cAxis[0]] = i;
				p1[cAxis[1]] = cLayer;
				p2[axis] = layer;
				p2[cAxis[0]] = this._dimensions[cAxis[0]] - i - 1;
				p2[cAxis[1]] = this._dimensions[cAxis[1]] - cLayer - 1;
				swaps.push([p1, p2]);
				if (p1[cAxis[1]] >= p2[cAxis[1]] && p1[cAxis[0]] == Math.floor(this._dimensions[cAxis[0]] / 2) - 1) {
					break layerLoop;
				}
			}
		}
		for (var i = 0; i < swaps.length; i++) {
			var saver = this._parts[swaps[i][0].x][swaps[i][0].y][swaps[i][0].z];
			this._parts[swaps[i][0].x][swaps[i][0].y][swaps[i][0].z] = this._parts[swaps[i][1].x][swaps[i][1].y][swaps[i][1].z];
			this._parts[swaps[i][1].x][swaps[i][1].y][swaps[i][1].z] = saver;
		}
	};
	RubiksCube.prototype.getPartIndex = function (part) {
		for (var i = 0; i < this._dimensions.x; i++) {
			for (var j = 0; j < this._dimensions.y; j++) {
				for (var k = 0; k < this._dimensions.z; k++) {
					if (part == this._parts[i][j][k]) {
						return new THREE.Vector3(i, j, k);
					}
				}
			}
		}
		return null;
	};
	RubiksCube.prototype.normalToFace = function (normal) {
		var result = "L";
		var dist = normal.angleTo(this._faces.L);
		for (var _i = 0, _a = ["U", "F", "D", "R", "B"]; _i < _a.length; _i++) {
			var face = _a[_i];
			var angle = normal.angleTo(this._faces[face]);
			if (angle < dist) {
				dist = angle;
				result = face;
			}
		}
		return result;
	};
	RubiksCube.prototype.angleToDirection = function (a) {
		if (Math.abs(a) > 135) {
			return 2;
		}
		if (a > 45) {
			return 1;
		}
		if (a < -45) {
			return 3;
		}
		return 0;
	};
	RubiksCube.prototype.touchToMove = function (partIndex, face, direction) {
		var move = this._faceAxis[face][direction];
		move = move.slice(0, 1) + partIndex[move.charAt(0).toLowerCase()] + move.slice(1);
		return move;
	};
	RubiksCube.prototype._translateMove = function (m) {
		var move = this._moveTranslator[m.charAt(0)];
		if (m.indexOf("'") > -1) {
			if (move.indexOf("'") > -1) {
				move = move.replace("'", "");
			} else {
				move += "'";
			}
		}
		if (m.indexOf("2") > -1) {
			move += "*";
		}
		return move;
	};
	RubiksCube.prototype._initMoveTranslator = function () {
		this._moveTranslator.U = "Y" + (this._dimensions.y - 1);
		this._moveTranslator.F = "Z" + (this._dimensions.z - 1);
		this._moveTranslator.R = "X" + (this._dimensions.x - 1);
		if (this._dimensions.x % 2 == 1) {
			this._moveTranslator.M = "X" + (this._dimensions.x - 1) / 2 + "'";
		} else {
			delete this._moveTranslator.M;
		}
		if (this._dimensions.y % 2 == 1) {
			this._moveTranslator.E = "Y" + (this._dimensions.y - 1) / 2 + "'";
		} else {
			delete this._moveTranslator.E;
		}
		if (this._dimensions.z % 2 == 1) {
			this._moveTranslator.S = "Z" + (this._dimensions.z - 1) / 2 + "'";
		} else {
			delete this._moveTranslator.S;
		}
	};
	return RubiksCube;
})(MovingPuzzle);
var RubiksCuboid = (function (_super) {
	__extends(RubiksCuboid, _super);

	function RubiksCuboid(dimensions, blockAxis) {
		_super.call(this);
		this._blockAxis = blockAxis;
		this._dimensions = dimensions;
	}
	return RubiksCuboid;
})(RubiksCube);
var RubiksCubeSolver = (function (_super) {
	__extends(RubiksCubeSolver, _super);

	function RubiksCubeSolver() {
		_super.call(this, new THREE.Vector3(3, 3, 3), []);
	}
	RubiksCubeSolver.prototype.init = function () {
		_super.prototype.init.call(this);
		var indexArr = [6, 33, 53, 7, 30, 8, 27, 24, 3, 50, 4, 5, 21, 0, 9, 47, 1, 12, 2, 15, 18, 34, 52, 31, 28, 25, 49, 22, 10, 46, 13, 16, 19, 44, 35, 51, 43, 32, 42, 29, 26, 41, 48, 40, 39, 23, 38, 11, 45, 37, 14, 36, 17, 20];
		var p = 0;
		for (var i = 0; i < this.children.length; i++) {
			for (var j = 0; j < this.children[i].children.length; j++) {
				this.children[i].children[j].material.color.setHex(16777215);
				this.children[i].children[j].tileIndex = indexArr[p];
				p++;
			}
		}
	};
	RubiksCubeSolver.prototype._getObjectsForIntersection = function () {
		var o = this.children.slice();
		for (var i = 0; i < this.children.length; i++) {
			for (var j = 0; j < this.children[i].children.length; j++) {
				o.push(this.children[i].children[j]);
			}
		}
		return o;
	};
	return RubiksCubeSolver;
})(RubiksCuboid);