var maxtime = 90;
var maxmoves = 30;
var strcube = null;
var worker = new Worker("./assets/js/CubeWorker.js");

var totalTime = 0;

worker.addEventListener("error", function (event) {
    // console.log(event);
    event.preventDefault();
}, false);

window.hasCanvas = function () {
    var elem = document.createElement('canvas');
    return !!(elem.getContext && elem.getContext('2d'));
};

window.hasWebGL = function () {
    try {
        var canvas = document.createElement('canvas');
        return !!window.WebGLRenderingContext && (
            canvas.getContext('webgl') || canvas.getContext('experimental-webgl'));
    } catch (e) { return false; }
};

if (systemLanguage == "tr") {
    var mExplanations = {
        "R": "SAĞ bölümü saat yönünde 90° çevirin.",
        "R'": "SAĞ bölümü saat yönünün tersine 90° çevirin.",
        "R2": "SAĞ bölümü 180° döndürün.",
        "L": "SOL bölümü saat yönünde 90° döndürün.",
        "L'": "SOL bölümü saatin tersi yönünde 90° döndürün.",
        "L2": "SOL bölümü 180° döndürün.",
        "U": "YUKARI bölümü saat yönünde 90° döndürün.",
        "U'": "YUKARI bölümü saat yönünün tersine 90° çevirin.",
        "U2": "YUKARI bölümü 180° döndürün.",
        "D": "AŞAĞI bölümü saat yönünde 90° döndürün.",
        "D'": "AŞAĞI bölümü saat yönünün tersine 90° çevirin.",
        "D2": "AŞAĞI bölümü 180° döndürün.",
        "F": "ÖN bölümü saat yönünde 90° döndürün.",
        "F'": "ÖN bölümü saat yönünün tersine 90° döndürün.",
        "F2": "ÖN bölümü 180° döndürün.",
        "B": "ARKA bölümü saat yönünde 90° döndürün.",
        "B'": "ARKA bölümü saat yönünün tersine 90° çevirin.",
        "B2": "ARKA bölümü 180° döndürün."
    }
} else {
    var mExplanations = {
        "R": "Turn the RIGHT layer 90&deg; clockwise.",
        "R'": "Turn the RIGHT layer 90&deg; counterclockwise.",
        "R2": "Turn the RIGHT layer 180&deg;.",
        "L": "Turn the LEFT layer 90&deg; clockwise.",
        "L'": "Turn the LEFT layer 90&deg; counterclockwise.",
        "L2": "Turn the LEFT layer 180&deg;.",
        "U": "Turn the TOP layer 90&deg; clockwise.",
        "U'": "Turn the TOP layer 90&deg; counterclockwise.",
        "U2": "Turn the TOP layer 180&deg;.",
        "D": "Turn the BOTTOM layer 90&deg; clockwise.",
        "D'": "Turn the BOTTOM layer 90&deg; counterclockwise.",
        "D2": "Turn the BOTTOM layer 180&deg;.",
        "F": "Turn the FRONT layer 90&deg; clockwise.",
        "F'": "Turn the FRONT layer 90&deg; counterclockwise.",
        "F2": "Turn the FRONT layer 180&deg;.",
        "B": "Turn the BACK layer 90&deg; clockwise.",
        "B'": "Turn the BACK layer 90&deg; counterclockwise.",
        "B2": "Turn the BACK layer 180&deg;."
    }
}

window.solver = {
    currentColor: 'D50000',
    puzzleState: [],
    steps: 0,
    currentStep: 0,
    solveSeq: [],

    renderer: null,
    scene: null,
    camera: null,
    puzzle: null,
    hammer: null,

    moveExplanations: mExplanations,

    resize: function () {
        var container = $('#puzzleContainer');

        var DPR = (window.devicePixelRatio) ? window.devicePixelRatio : 1;
        window.solver.renderer.setPixelRatio(DPR);
        var minLength = Math.min(container[0].offsetWidth, container[0].offsetHeight)
        window.solver.renderer.setSize(container[0].offsetWidth, container[0].offsetHeight);
        window.solver.camera.aspect = container[0].offsetWidth / container[0].offsetHeight;
        window.solver.camera.updateProjectionMatrix();
    },

    render: function () {
        window.solver.renderer.render(window.solver.scene, window.solver.camera);
        requestAnimationFrame(window.solver.render);
    },

    reset: function () {
        var objects = window.solver.puzzle._getObjectsForIntersection();
        for (var i = 0; i < objects.length; i++)
            if (objects[i].hasOwnProperty('isTile'))
                objects[i].material.color.setHex(0xFFFFFF);
    },

    init: function () {
        var container = document.getElementById("puzzleContainer");
        var $container = $(container);
        loadingShow();
        $container.find('.menu').not('#solveInstructions').addClass('shown');

        if (window.hasWebGL())
            window.solver.renderer = new THREE.WebGLRenderer({ antialias: true, alpha: true });
        else
            window.solver.renderer = new THREE.CanvasRenderer({ alpha: true });

        window.solver.renderer.setClearColor(0xffffff, 0);
        window.solver.scene = new THREE.Scene();

        var light = new THREE.DirectionalLight(0xffffff, 1.7);
        light.position.set(5, 5, 5);
        window.solver.scene.add(light);

        window.solver.camera = new THREE.PerspectiveCamera(45, container.offsetWidth / container.offsetHeight, 1, 1000);
        window.solver.camera.position.set(6.5, 5.5, 6.5);
        window.solver.camera.lookAt(new THREE.Vector3(0, 0, 0));

        window.solver.puzzle = new RubiksCubeSolver();
        window.solver.puzzle.init();
        window.solver.puzzle.position.y = -.2;
        window.solver.scene.add(window.solver.puzzle);

        container.appendChild(window.solver.renderer.domElement);

        window.solver.resize();
        window.solver.render();

        window.solver.hammer = new Hammer(window.solver.renderer.domElement, {
            drag_block_vertical: true,
            drag_lock_to_axis: true
        });
        window.solver.hammer.get('swipe').set({ direction: Hammer.DIRECTION_ALL });

        window.solver.addTouch();

        $(window).on('resize', function () {
            window.solver.resize();
        });
        setTimeout(function () {
            $('#fullScreenBtn').click();
            loadingHide();
            worker.postMessage({ type: "generateTables" });
        }, 555);
    },

    addTouch: function () {
        window.solver.hammer.on('swipeleft', function (e) {
            window.solver.puzzle.addMove('Y');
        });
        window.solver.hammer.on('swiperight', function (e) {
            window.solver.puzzle.addMove('Y\'');
        });
        window.solver.hammer.on('swipeup', function (e) {
            var offset = $(window.solver.renderer.domElement).offset();
            var x = e.center.x - offset.left;
            var width = e.target.offsetWidth;
            if ((2 * x) > width)
                window.solver.puzzle.addMove('Z\'');
            else
                window.solver.puzzle.addMove('X');
        });
        window.solver.hammer.on('swipedown', function (e) {
            var offset = $(window.solver.renderer.domElement).offset();
            var x = e.center.x - offset.left;
            var width = e.target.offsetWidth;
            if ((2 * x) > width)
                window.solver.puzzle.addMove('Z');
            else
                window.solver.puzzle.addMove('X\'');
        });

        window.solver.hammer.on('tap', function (e) {
            var offset = $(window.solver.renderer.domElement).offset();
            var x = e.center.x - offset.left;
            var y = e.center.y - offset.top;

            var touch = window.solver.puzzle.isPointInside(x, y, window.solver.camera, window.solver.renderer.domElement);

            if (touch && (touch.object.hasOwnProperty('isTile'))) {
                touch.object.material.color.setHex(window.solver.currentColor);
            }
        });
    },

    removeTouch: function () {
        window.solver.hammer.off('swipeup');
        window.solver.hammer.off('swipedown');
        window.solver.hammer.off('swipeleft');
        window.solver.hammer.off('swiperight');
        window.solver.hammer.off('tap');
    },

    getState: function () {
        var colors = [
            0x2ABB2A, 0xFFFFFF, 0xC7242F,
            0xFFDD00, 0x3579DE, 0xFF6D00
        ];

        var objects = [];
        for (var i = 0; i < window.solver.puzzle._parts.length; i++)
            for (var j = 0; j < window.solver.puzzle._parts[i].length; j++)
                for (var k = 0; k < window.solver.puzzle._parts[i][j].length; k++)
                    if (window.solver.puzzle._parts[i][j][k] != null)
                        objects.push(window.solver.puzzle._parts[i][j][k]);

        var converter = [
            { "X": 6, "Y": 33, "Z": 53 }, { "X": 7, "Y": 30 }, { "X": 8, "Y": 27, "Z": 24 },
            { "X": 3, "Z": 50 }, { "X": 4 }, { "X": 5, "Z": 21 },
            { "X": 0, "Y": 9, "Z": 47 }, { "X": 1, "Y": 12 }, { "X": 2, "Y": 15, "Z": 18 },
            { "Y": 34, "Z": 52 }, { "Y": 31 }, { "Y": 28, "Z": 25 },
            { "Z": 49 }, { "Z": 22 },
            { "Y": 10, "Z": 46 }, { "Y": 13 }, { "Y": 16, "Z": 19 },
            { "X": 44, "Y": 35, "Z": 51 }, { "X": 43, "Y": 32 }, { "X": 42, "Y": 29, "Z": 26 },
            { "X": 41, "Z": 48 }, { "X": 40 }, { "X": 39, "Z": 23 },
            { "X": 38, "Y": 11, "Z": 45 }, { "X": 37, "Y": 14 }, { "X": 36, "Y": 17, "Z": 20 }
        ];

        var faces = {
            X: new THREE.Vector3(1, 0, 0),
            Y: new THREE.Vector3(0, 1, 0),
            Z: new THREE.Vector3(0, 0, 1)
        };

        var tiles = new Array(24);
        for (var i = 0; i < objects.length; i++) {
            var part = objects[i];
            for (var j = 0; j < part.children.length; j++) {
                var color = colors.indexOf(part.children[j].material.color.getHex());
                var normal = part.children[j].geometry.faces[0].normal.clone().applyMatrix4(
                    new THREE.Matrix4().extractRotation(part.children[j].matrixWorld.clone())
                );

                var face = 'X';

                var angle = normal.angleTo(faces.Y) % Math.PI;
                angle = Math.min(
                    Math.abs(angle),
                    Math.abs(angle - Math.PI)
                );
                if (angle < 0.1)
                    face = 'Y';
                else {
                    angle = normal.angleTo(faces.Z) % Math.PI;
                    angle = Math.min(
                        Math.abs(angle),
                        Math.abs(angle - Math.PI)
                    );
                    if (angle < 0.1)
                        face = 'Z';
                }

                tiles[converter[i][face]] = color;
            }
        }

        return tiles;
    },
    getState2: function () {
        var colors = [
            0x2ABB2A, 0xFFFFFF, 0xC7242F,
            0xFFDD00, 0x3579DE, 0xFF6D00
        ];

        tiles = new Array(24);

        for (var i = 0; i < window.solver.puzzle.children.length; i++)
            for (var j = 0; j < window.solver.puzzle.children[i].children.length; j++) {
                var tile = window.solver.puzzle.children[i].children[j];
                var color = colors.indexOf(tile.material.color.getHex());
                tiles[tile.tileIndex] = color;
            }

        return tiles;
    },

    solve: function () {
        var $state = window.solver.puzzleState.join('');
        // console.log($state);
        loadingShow();

        if (runAdmob != null) {
            runAdmob.showInterstitial();
        }

        var raw = $state;
        var sol = raw.slice(0, 9);
        var ust = raw.slice(9, 18);
        var on = raw.slice(18, 27);
        var alt = raw.slice(27, 36);
        var sag = raw.slice(36, 45);
        var arka = raw.slice(45, 54);
        var strcubejs = ust + sag + on + alt + sol + arka;
        var U = ust.slice(4, 5);
        var R = sag.slice(4, 5);
        var F = on.slice(4, 5);
        var D = alt.slice(4, 5);
        var L = sol.slice(4, 5);
        var B = arka.slice(4, 5);
        //_ _ _ _4U_ _ _ _ _ _ _ _13R_ _ _ _ _ _ _ _22F_ _ _ _ _ _ _ _31D_ _ _ _ _ _ _ _40L_ _ _ _ _ _ _ _49B_ _ _ _
        // 5 turuncu      F
        // 4 mavi         U
        // 3 sarı         B
        // 2 kırmızı      R
        // 1 beyaz        D
        // 0 yeşil        L

        // U R F D L B
        // üst  sağ ön alt   sol arka --cubejs

        //sol   üst  ön  alt  sağ  arka --grubiks
        function numberToString(n) {
            switch (n) {
                case U:
                    return "U";
                case R:
                    return "R";
                case F:
                    return "F";
                case D:
                    return "D";
                case L:
                    return "L";
                case B:
                    return "B";
            }
        }
        var newStrCubejs = numberToString(strcubejs.slice(0, 1));
        for (i = 1; i < strcubejs.length; i++) {
            newStrCubejs += numberToString(strcubejs.slice(i, i + 1));
        }

        strcube = newStrCubejs;
        worker.postMessage({ type: "verify", cube: strcube });
        // console.log(newStrCubejs);
    },

    preCalcTests: function () {
        var error = '';

        var counters = [];
        $.each(['green', 'white', 'red', 'yellow', 'blue', 'orange'], function (key, value) {
            counters.push({ name: value, count: 0 });
        });

        for (var i = 0; i < window.solver.puzzleState.length; i++)
            counters[window.solver.puzzleState[i]].count++;

        for (var i = 0; i < counters.length; i++)
            if (counters[i].count < 9){
                if (systemLanguage == "tr") {
                    error += '<li><div class="color-div" style="background-color:' + counters[i].name + '"></div> alan sayısı eksik.</li>';
                }else{
                    error += '<li>You do not have enough <div class="color-div" style="background-color:' + counters[i].name + '"></div> tiles.</li>';
                }
            } 
            else if (counters[i].count > 9){
                if (systemLanguage == "tr") {
                    error += '<li><div class="color-div" style="background-color:' + counters[i].name + '"></div> alan sayısı fazla.</li>';
                }else{
                    error += '<li>You have too many <div class="color-div" style="background-color:' + counters[i].name + '"></div> tiles.</li>';
                }
            }

        if (error != '') throw error;

        var isSolved = true;
        for (var i = 1; i < 9; i++)
            if (
                (window.solver.puzzleState[0] != window.solver.puzzleState[i]) ||
                (window.solver.puzzleState[9] != window.solver.puzzleState[9 + i]) ||
                (window.solver.puzzleState[18] != window.solver.puzzleState[18 + i]) ||
                (window.solver.puzzleState[27] != window.solver.puzzleState[27 + i]) ||
                (window.solver.puzzleState[36] != window.solver.puzzleState[36 + i])
            ) isSolved = false;
        if (isSolved){
            if (systemLanguage == "tr") {
                throw error = '<li>Küp zaten çözülü! Karışık bir küp üzerinde tekrar deneyiniz :)</li>';
            }
            else{
                throw error = '<li>Your cube is already SOLVED! :)</li>';
            }
        }
            
    }
};

function disableDocTouch(e) {
    e.preventDefault();
}

worker.addEventListener("message", function (event) {
    if (event.data.type == "message") {
        // console.log(event.data.msg);
    } else if (event.data.type == "verify") {
        if (event.data.result == 0) {
            // errMessage("Valid cube.")
            // console.log("Valid cube.");
            letSolve();
        } else {
            var message;
            switch (event.data.result) {
                case -1:
                    message = "There is not exactly one face of each color";
                    break;
                case -2:
                    message = "Not all 12 edges exist exactly once";
                    break;
                case -3:
                    message = "Flip error: One edge has to be flipped";
                    break;
                case -4:
                    message = "Not all corners exist exactly once";
                    break;
                case -5:
                    message = "Twist error: One corner has to be twisted";
                    break;
                case -6:
                    message = "Parity error: Two corners or two edges have to be exchanged";
                    break;
                default:
                    message = "Unknown error";
                    break;
            }
            if (systemLanguage == "tr") {
                errMessage("Küp çözme işlemi sırasında hatalar tespit edildi, lütfen küp renklendirmelerini kontrol ederek tekrar deneyiniz!");
            }
            else{
                errMessage("Cube invalid. Error " + event.data.result + ": " + message + ".")
            }
        }
    } else if (event.data.type == "progress") {
        // console.log(event.data.type);
    } else if (event.data.type == "solution") {
        if (event.data.result.indexOf("Error") == 0) {
            var message;
            var err = parseInt(event.data.result.substring(6, 7));
            switch (err) {
                case 1:
                    message = "There is not exactly one facelet of each color";
                    break;
                case 2:
                    message = "Not all 12 edges exist exactly once";
                    break;
                case 3:
                    message = "Flip error: One edge has to be flipped";
                    break;
                case 4:
                    message = "Not all corners exist exactly once";
                    break;
                case 5:
                    message = "Twist error: One corner has to be twisted";
                    break;
                case 6:
                    message = "Parity error: Two corners or two edges have to be exchanged";
                    break;
                case 7:
                    message = "No solution exists in " + maxmoves + " number of moves or less";
                    break;
                case 8:
                    message = "Timed out; No solution found within " + maxtime + " seconds";
                    break;
                default:
                    message = "Unknown error";
                    break;
            }
            if (systemLanguage == "tr") {
                errMessage("Küp çözme işlemi sırasında hatalar tespit edildi, lütfen küp renklendirmelerini kontrol ederek tekrar deneyiniz!");
            }
            else{
                errMessage((err < 7 ? "Invalid Cube. " : "Solver ") + "Error -" + err + ": " + message + ".");
            }
        } else {
            var response = {};
            // console.log(event.data.result);
            response.moves = event.data.result;
            window.tRes = response;
            // console.log(response);
            if (!Array.isArray(response.moves)) {
                response.moves = $.trim(response.moves.replace(/\s/g, ' '));
                response.moves = response.moves.split(' ');
            }
            solveReady({ moves: response.moves.join(' ') });
        }
    }
}, false);

function letSolve() {
    worker.postMessage({ type: "solve", cube: strcube, maxDepth: maxmoves, maxTime: maxtime });
}

function loadingShow(){
    if($('#loading').hasClass("dpn")){
        $('#loading').removeClass("dpn");
    }
}

function loadingHide(){
    if(!$('#loading').hasClass("dpn")){
        $('#loading').addClass("dpn");
    }
}

function errMessage(msg) {
    loadingHide();
    if (systemLanguage == "tr") {
        $('#coloringErrorDetails').html('<li>Küp çözülemedi<br><strong>Not:</strong>' + msg + '<br> Küpünüzü parçalara ayırdıysanız, yanlış birleştirmiş olabilirsiniz.</li>');
        $('#coloringErrorModal').modal();
        $('#solveBtn').prop('disabled', false).text('Başla');
    }
    else{
        $('#coloringErrorDetails').html('<li>Your cube cannot be solved.<br><strong>Note:</strong>' + msg + '<br> if you\'ve taken your cube apart you might have reassembled it incorrectly</li>');
        $('#coloringErrorModal').modal();
        $('#solveBtn').prop('disabled', false).text('Solve!');
    }
}

function solveReady(obj) {
    
    $('#resetPuzzleModalBtn').fadeOut();
    loadingHide();
    window.solver.puzzle.addListener('layerAnimationComplete', function () {
        if (window.solver.puzzle._animation.moveQueue.length)
            $('#nextBtn').prop('disabled', false);
        else {
            if (runAdmob != null) {
                runAdmob.showInterstitial();
            }
            if (systemLanguage == "tr") {
                $('#nextBtn').html('İşlem tamam!');
            }
            else{
                $('#nextBtn').html('Done!');
            }
        }
    });
    window.solver.puzzle._animation.autoplay = false;
    var $moves = $.trim(obj.moves.replace(/  /g, ' '));
    $moves = $moves.split(' ');
    window.solver.puzzle._animation.moveQueue = $moves.slice();
    window.solver.currentStep = 0;
    window.solver.steps = $moves.length;
    window.solver.solveSeq = $moves.slice(0);
    window.solver.removeTouch();
    $('#solveBtn').fadeOut();
    $('#colorPalette').removeClass('shown');
    $('#solveInstructions').addClass('shown');
}

window.domReadyQueue.push(function ($) {
    $('.colorPicker').each(function (index, value) {
        var $this = $(value);
        $this.css('background', $this.data('color'));
    }).on('click', function () {
        $('.colorPicker').removeClass('selected');
        var $this = $(this).addClass('selected');
        window.solver.currentColor = Number($this.data('color').replace('#', '0x'));
    }).eq(0).click();

    $(document).on('touchmove', disableDocTouch);
    $('#fullScreenBtn').on('click', function () {
        var $cont = $('#puzzleContainer');
        $cont.toggleClass('fullScreen');

        if ($cont.hasClass('fullScreen'))
            $(document).on('touchmove', disableDocTouch);
        else
            $(document).off('touchmove', disableDocTouch);

        window.solver.resize();
    });

    $('#resetPuzzleBtn').on('click', function () {
        if (runAdmob != null) {
            runAdmob.showInterstitial();
        }
        window.solver.reset();
    });

    $('#nextBtn').on('click', function () {
        $(this).prop('disabled', true);
        window.solver.currentStep++;
        if (systemLanguage == "tr") {
            $('#stepHeading').text('İlerleme durumu: ' + window.solver.currentStep + ' - ' + window.solver.steps);
        }
        else{
            $('#stepHeading').text('Step ' + window.solver.currentStep + ' of ' + window.solver.steps);
        }
        $('#moveExplanation').html(window.solver.moveExplanations[window.solver.solveSeq[window.solver.currentStep - 1]]);
        window.solver.puzzle.execMoveQueue();
    });

    $('#solveBtn').on('click', function () {
        var $btn = $(this);
        $('#resetPuzzleModalBtn').prop('disabled', true);
        if (systemLanguage == "tr") {
            $btn.prop('disabled', true).text('...');
        }
        else{
            $btn.prop('disabled', true).text('Solving...');
        }
        window.solver.puzzleState = window.solver.getState();
        try {
            window.solver.preCalcTests();
        } catch (e) {
            $('#coloringErrorDetails').html(e);
            $('#coloringErrorModal').modal();
            $('#resetPuzzleModalBtn').prop('disabled', false);
            if (systemLanguage == "tr") {
                $btn.prop('disabled', false).text('Başla');
            }
            else{
                $btn.prop('disabled', false).text('Solve!');
            }
            return false;
        }

        window.solver.solve();
    });

    $('#exitBtn').on('click', function () {
        if (closeEvent != null) {
            closeEvent.exitCubeSolver();
        }
    });
    window.solver.init();
});
